@@ -1,6 +1,7 @@
 from math import sqrt, cos, sin, acos, degrees, radians, log, pi
 from bisect import bisect
 from abc import ABC, abstractmethod
+import math
 
 try:
     from collections.abc import MutableSequence
@@ -14,6 +15,61 @@ MIN_DEPTH = 5
 ERROR = 1e-12
 
 
+def _find_solutions_for_bezier(c2, c1, c0):
+    """Find solutions of c2 * t^2 + c1 * t + c0 = 0 where t in [0, 1]"""
+    soln = []
+    if c2 == 0:
+        if c1 != 0:
+            soln.append(-c0 / c1)
+    else:
+        det = c1**2 - 4 * c2 * c0
+        if det >= 0:
+            soln.append((-c1 + math.pow(det, 0.5)) / 2.0 / c2)
+            soln.append((-c1 - math.pow(det, 0.5)) / 2.0 / c2)
+    return [s for s in soln if 0.0 <= s and s <= 1.0]
+
+
+def _find_solutions_for_arc(a, b, c, d):
+    """Find solution for a sin(x) + b cos(x) = 0 where x = c + d * t and t in [0, 1]"""
+    if a == 0:
+        # when n \in Z
+        # pi / 2 + pi * n = c + d * t
+        # --> n = d / pi * t - (1/2 - c/pi)
+        # --> t = (pi / 2 - c + pi * n) / d
+        n_ranges = [-0.5 + c / math.pi, d / math.pi - 0.5 + c / math.pi]
+        n_range_start = math.floor(min(n_ranges))
+        n_range_end = math.ceil(max(n_ranges))
+        t_list = [
+            (math.pi / 2 - c + math.pi * n) / d
+            for n in range(n_range_start, n_range_end + 1)
+        ]
+    elif b == 0:
+        # when n \in Z
+        # pi * n = c + d * t
+        # --> n = d / pi * t + c / pi
+        # --> t = (- c + pi * n) / d
+        n_ranges = [c / math.pi, d / math.pi + c / math.pi]
+        n_range_start = math.floor(min(n_ranges))
+        n_range_end = math.ceil(max(n_ranges))
+        t_list = [(-c + math.pi * n) / d for n in range(n_range_start, n_range_end + 1)]
+    else:
+        # when n \in Z
+        # arct = tan^-1 (- b / a)  and
+        # arct + pi * n = c + d * t
+        # --> n = (c - arct + d * t) / pi
+        # --> t = (arct - c + pi * n) / d
+        arct = math.atan(-b / a)
+        n_ranges = [(c - arct) / math.pi, d / math.pi + (c - arct) / math.pi]
+        n_range_start = math.floor(min(n_ranges))
+        n_range_end = math.ceil(max(n_ranges))
+        t_list = [
+            (arct - c + math.pi * n) / d for n in range(n_range_start, n_range_end + 1)
+        ]
+
+    t_list = [t for t in t_list if 0.0 <= t and t <= 1.0]
+    return t_list
+
+
 def segment_length(curve, start, end, start_point, end_point, error, min_depth, depth):
     """Recursively approximates the length by straight lines"""
     mid = (start + end) / 2
@@ -57,6 +113,10 @@ class PathSegment(ABC):
         number of iterations.
         """
 
+    @abstractmethod
+    def boundingbox(self):
+        """Returns the bounding box of a path in the format of [left, top, right, bottom]"""
+
 
 class NonLinear(PathSegment):
     """A line that is not straight
@@ -118,7 +178,7 @@ class Line(Linear):
 
         if self.horizontal and self.is_horizontal_from(previous):
             cmd = "h" if self.relative else "H"
-            return f"{cmd} {x:G},{y:G}"
+            return f"{cmd} {x:G}"
 
         if self.vertical and self.is_vertical_from(previous):
             cmd = "v" if self.relative else "V"
@@ -133,6 +193,13 @@ class Line(Linear):
     def is_horizontal_from(self, previous):
         return self.start == previous.end and self.start.imag == self.end.imag
 
+    def boundingbox(self):
+        x_min = min(self.start.real, self.end.real)
+        x_max = max(self.start.real, self.end.real)
+        y_min = min(self.start.imag, self.end.imag)
+        y_max = max(self.start.imag, self.end.imag)
+        return [x_min, y_min, x_max, y_max]
+
 
 class CubicBezier(NonLinear):
     def __init__(self, start, control1, control2, end, relative=False, smooth=False):
@@ -221,6 +288,41 @@ class CubicBezier(NonLinear):
         end_point = self.point(1)
         return segment_length(self, 0, 1, start_point, end_point, error, min_depth, 0)
 
+    def boundingbox(self):
+        """Calculate the bounding box of a cubic Bezier curve.
+
+        A cubic Bezier curve and its derivative are given as follows.
+
+        P(t)  = (1-t)^3 P0 + 3 t (1-t)^2 P1 + 3 t^2 (1-t) P2 + t^3 P_3
+        P'(t) = 3(1-t)^2 (P1-P0) + 6(1-t)t (P2-P1) + 3 t^2 (P3 - P2)
+        """
+        g0 = self.control1 - self.start
+        g1 = self.control2 - self.control1
+        g2 = self.end - self.control2
+
+        c0 = 3 * g0
+        c1 = -6 * g0 + 6 * g1
+        c2 = 3 * g0 - 6 * g1 + 3 * g2
+
+        x_c0, x_c1, x_c2 = [c.real for c in [c0, c1, c2]]
+        y_c0, y_c1, y_c2 = [c.imag for c in [c0, c1, c2]]
+
+        x_cand = [0, 1] + _find_solutions_for_bezier(x_c2, x_c1, x_c0)
+        y_cand = [0, 1] + _find_solutions_for_bezier(y_c2, y_c1, y_c0)
+
+        x_coords = []
+        y_coords = []
+        for t in x_cand:
+            p = self.point(t)
+            x_coords.append(p.real)
+        for t in y_cand:
+            p = self.point(t)
+            y_coords.append(p.imag)
+
+        x_min, x_max = min(x_coords), max(x_coords)
+        y_min, y_max = min(y_coords), max(y_coords)
+        return [x_min, y_min, x_max, y_max]
+
 
 class QuadraticBezier(NonLinear):
     def __init__(self, start, control, end, relative=False, smooth=False):
@@ -326,6 +428,39 @@ class QuadraticBezier(NonLinear):
                     s = abs(a) * (k**2 / 2 - k + 1)
         return s
 
+    def boundingbox(self):
+        """Calculate the bounding box of a quadratic Bezier curve.
+
+        A quadratic Bezier curve and its derivative are given as follows.
+
+        P(t)  = (1-t)^2 P0 + 2 t (1-t)^2 P1 + t^2 P2
+        P'(t) = 2(1-t) (P1-P0) + 2t (P2-P1)
+        """
+        g0 = self.control - self.start
+        g1 = self.end - self.control
+
+        c0 = 2 * g0
+        c1 = -2 * g0 + 2 * g1
+
+        x_c0, x_c1 = [c.real for c in [c0, c1]]
+        y_c0, y_c1 = [c.imag for c in [c0, c1]]
+
+        x_cand = [0, 1] + _find_solutions_for_bezier(0, x_c1, x_c0)
+        y_cand = [0, 1] + _find_solutions_for_bezier(0, y_c1, y_c0)
+
+        x_coords = []
+        y_coords = []
+        for t in x_cand:
+            p = self.point(t)
+            x_coords.append(p.real)
+        for t in y_cand:
+            p = self.point(t)
+            y_coords.append(p.imag)
+
+        x_min, x_max = min(x_coords), max(x_coords)
+        y_min, y_max = min(y_coords), max(y_coords)
+        return [x_min, y_min, x_max, y_max]
+
 
 class Arc(NonLinear):
     def __init__(self, start, radius, rotation, arc, sweep, end, relative=False):
@@ -515,6 +650,53 @@ class Arc(NonLinear):
         end_point = self.point(1)
         return segment_length(self, 0, 1, start_point, end_point, error, min_depth, 0)
 
+    def boundingbox(self):
+        """Calculate the bounding box of an arc
+
+        To calculate the extremums of the arc coordinates, we solve
+
+        x'(angle) = - cosr * radius.real * sin(angle)
+                    - sinr * radius.imag * cos(angle) = 0
+
+        y'(angle) = - sinr * radius.real * sin(angle)
+                    + cosr * radius.imag * cos(angle) = 0
+
+        and
+            angle = radians(self.theta + (self.delta * pos))
+        where pos ranges from 0 to 1
+        """
+
+        # angle = radians(self.theta + (self.delta * pos))
+        cosr = cos(radians(self.rotation))
+        sinr = sin(radians(self.rotation))
+        radius = self.radius * self.radius_scale
+
+        x_a = -cosr * radius.real
+        x_b = -sinr * radius.imag
+        x_c = radians(self.theta)
+        x_d = radians(self.delta)
+
+        y_a = -sinr * radius.real
+        y_b = +cosr * radius.imag
+        y_c = radians(self.theta)
+        y_d = radians(self.delta)
+
+        x_pos = [0, 1.0] + _find_solutions_for_arc(x_a, x_b, x_c, x_d)
+        y_pos = [0, 1.0] + _find_solutions_for_arc(y_a, y_b, y_c, y_d)
+
+        x_coords = []
+        y_coords = []
+        for pos in x_pos:
+            p = self.point(pos)
+            x_coords.append(p.real)
+        for pos in y_pos:
+            p = self.point(pos)
+            y_coords.append(p.imag)
+
+        x_min, x_max = min(x_coords), max(x_coords)
+        y_min, y_max = min(y_coords), max(y_coords)
+        return [x_min, y_min, x_max, y_max]
+
 
 class Move:
     """Represents move commands. Does nothing, but is there to handle
@@ -558,6 +740,13 @@ class Move:
     def length(self, error=ERROR, min_depth=MIN_DEPTH):
         return 0
 
+    def boundingbox(self):
+        x_min = min(self.start.real, self.end.real)
+        x_max = max(self.start.real, self.end.real)
+        y_min = min(self.start.imag, self.end.imag)
+        y_max = max(self.start.imag, self.end.imag)
+        return [x_min, y_min, x_max, y_max]
+
 
 class Close(Linear):
     """Represents the closepath command"""
@@ -573,6 +762,13 @@ class Close(Linear):
     def _d(self, previous):
         return "z" if self.relative else "Z"
 
+    def boundingbox(self):
+        x_min = min(self.start.real, self.end.real)
+        x_max = max(self.start.real, self.end.real)
+        y_min = min(self.start.imag, self.end.imag)
+        y_max = max(self.start.imag, self.end.imag)
+        return [x_min, y_min, x_max, y_max]
+
 
 class Path(MutableSequence):
     """A Path is a sequence of path segments"""
@@ -611,7 +807,6 @@ class Path(MutableSequence):
         return "Path(%s)" % (", ".join(repr(x) for x in self._segments))
 
     def __eq__(self, other):
-
         if not isinstance(other, Path):
             return NotImplemented
         if len(self) != len(other):
@@ -688,3 +883,18 @@ class Path(MutableSequence):
             previous_segment = segment
 
         return " ".join(parts)
+
+    def boundingbox(self):
+        x_coords = []
+        y_coords = []
+
+        for e in self:
+            x_min, y_min, x_max, y_max = e.boundingbox()
+            x_coords.append(x_min)
+            x_coords.append(x_max)
+            y_coords.append(y_min)
+            y_coords.append(y_max)
+
+        x_min, x_max = min(x_coords), max(x_coords)
+        y_min, y_max = min(y_coords), max(y_coords)
+        return [x_min, y_min, x_max, y_max]
